__author__ = "Simon Nilsson"

import glob
import itertools
import json
import logging
import logging.config
import os
import shutil
from ast import literal_eval
from configparser import ConfigParser
from datetime import datetime
from typing import Any, Dict, List, Optional, Tuple, Union

import cv2
import pandas as pd

try:
    from typing import Literal
except:
    from typing_extensions import Literal

from simba.utils.checks import (check_file_exist_and_readable, check_float,
                                check_if_dir_exists,
                                check_if_filepath_list_is_empty)
from simba.utils.enums import ConfigKey, Defaults, Dtypes, Keys, Paths
from simba.utils.errors import (BodypartColumnNotFoundError, DataHeaderError,
                                DuplicationError, InvalidInputError,
                                MissingProjectConfigEntryError,
                                NoFilesFoundError, NoROIDataError,
                                NotDirectoryError, ParametersFileError,
                                PermissionError)
from simba.utils.lookups import (create_color_palettes, get_color_dict,
                                 get_emojis, get_log_config)
from simba.utils.printing import SimbaTimer, stdout_success
from simba.utils.read_write import (find_core_cnt, get_all_clf_names,
                                    get_fn_ext, read_config_file, read_df,
                                    read_project_path_and_file_type, write_df)
from simba.utils.warnings import (BodypartColumnNotFoundWarning,
                                  InvalidValueWarning, NoDataFoundWarning,
                                  NoFileFoundWarning)


class ConfigReader(object):
    """
    Methods for reading SimBA configparser.Configparser project config and associated project data.

    :param configparser.Configparser config_path: path to SimBA project_config.ini
    :param bool read_video_info: if true, read the project_folder/logs/video_info.csv file.
    """

    def __init__(
        self, config_path: str, read_video_info: bool = True, create_logger: bool = True
    ):

        self.timer = SimbaTimer(start=True)
        self.config_path = config_path
        self.config = read_config_file(config_path=config_path)
        self.datetime = datetime.now().strftime("%Y%m%d%H%M%S")
        self.project_path, self.file_type = read_project_path_and_file_type(
            config=self.config
        )
        self.input_csv_dir = os.path.join(self.project_path, Paths.INPUT_CSV.value)
        self.features_dir = os.path.join(
            self.project_path, Paths.FEATURES_EXTRACTED_DIR.value
        )
        self.targets_folder = os.path.join(
            self.project_path, Paths.TARGETS_INSERTED_DIR.value
        )
        self.input_frames_dir = os.path.join(
            self.project_path, Paths.INPUT_FRAMES_DIR.value
        )
        self.machine_results_dir = os.path.join(
            self.project_path, Paths.MACHINE_RESULTS_DIR.value
        )
        self.directionality_df_dir = os.path.join(
            self.project_path, Paths.DIRECTIONALITY_DF_DIR.value
        )
        self.body_part_directionality_df_dir = os.path.join(
            self.project_path, Paths.BODY_PART_DIRECTIONALITY_DF_DIR.value
        )
        self.outlier_corrected_dir = os.path.join(
            self.project_path, Paths.OUTLIER_CORRECTED.value
        )
        self.outlier_corrected_movement_dir = os.path.join(
            self.project_path, Paths.OUTLIER_CORRECTED_MOVEMENT.value
        )
        self.heatmap_clf_location_dir = os.path.join(
            self.project_path, Paths.HEATMAP_CLF_LOCATION_DIR.value
        )
        self.heatmap_location_dir = os.path.join(
            self.project_path, Paths.HEATMAP_LOCATION_DIR.value
        )
        self.line_plot_dir = os.path.join(self.project_path, Paths.LINE_PLOT_DIR.value)
        self.probability_plot_dir = os.path.join(
            self.project_path, Paths.PROBABILITY_PLOTS_DIR.value
        )
        self.gantt_plot_dir = os.path.join(
            self.project_path, Paths.GANTT_PLOT_DIR.value
        )
        self.path_plot_dir = os.path.join(self.project_path, Paths.PATH_PLOT_DIR.value)
        self.shap_logs_path = os.path.join(self.project_path, Paths.SHAP_LOGS.value)
        self.video_info_path = os.path.join(self.project_path, Paths.VIDEO_INFO.value)
        self.frames_output_dir = os.path.join(
            self.project_path, Paths.FRAMES_OUTPUT_DIR.value
        )
        self.font = cv2.FONT_HERSHEY_COMPLEX
        self.roi_features_save_dir = os.path.join(
            self.project_path, Paths.ROI_FEATURES.value
        )
        self.configs_meta_dir = os.path.join(self.project_path, "configs")
        self.sklearn_plot_dir = os.path.join(
            self.project_path, Paths.SKLEARN_RESULTS.value
        )
        self.detailed_roi_data_dir = os.path.join(
            self.project_path, Paths.DETAILED_ROI_DATA_DIR.value
        )
        self.directing_animals_video_output_path = os.path.join(
            self.project_path, Paths.DIRECTING_BETWEEN_ANIMALS_OUTPUT_PATH.value
        )
        self.directing_body_part_animal_video_output_path = os.path.join(
            self.project_path,
            Paths.DIRECTING_BETWEEN_ANIMAL_BODY_PART_OUTPUT_PATH.value,
        )
        self.animal_cnt = self.read_config_entry(
            config=self.config,
            section=ConfigKey.GENERAL_SETTINGS.value,
            option=ConfigKey.ANIMAL_CNT.value,
            data_type=Dtypes.INT.value,
        )
        self.bodypart_direction = self.read_config_entry(
            self.config,
            section=ConfigKey.DIRECTIONALITY_SETTINGS.value,
            option=ConfigKey.BODYPART_DIRECTION_VALUE.value,
            data_type=Dtypes.STR.value,
            default_value=Dtypes.NONE.value,
        )
        self.clf_cnt = self.read_config_entry(
            self.config,
            ConfigKey.SML_SETTINGS.value,
            ConfigKey.TARGET_CNT.value,
            Dtypes.INT.value,
        )
        self.clf_names = get_all_clf_names(config=self.config, target_cnt=self.clf_cnt)
        self.feature_file_paths = glob.glob(self.features_dir + f"/*.{self.file_type}")
        self.target_file_paths = glob.glob(self.targets_folder + f"/*.{self.file_type}")
        self.input_csv_paths = glob.glob(self.input_csv_dir + f"/*.{self.file_type}")
        self.body_part_directionality_paths = glob.glob(
            self.body_part_directionality_df_dir + f"/*.{self.file_type}"
        )
        self.outlier_corrected_paths = glob.glob(
            self.outlier_corrected_dir + f"/*.{self.file_type}"
        )
        self.outlier_corrected_movement_paths = glob.glob(
            self.outlier_corrected_movement_dir + f"/*.{self.file_type}"
        )
        self.cpu_cnt, self.cpu_to_use = find_core_cnt()
        self.machine_results_paths = glob.glob(
            self.machine_results_dir + f"/*.{self.file_type}"
        )
        self.logs_path = os.path.join(self.project_path, "logs")
        self.body_parts_path = os.path.join(self.project_path, Paths.BP_NAMES.value)
        check_file_exist_and_readable(file_path=self.body_parts_path)
        self.body_parts_lst = (
            pd.read_csv(self.body_parts_path, header=None).iloc[:, 0].to_list()
        )
        self.body_parts_lst = [x for x in self.body_parts_lst if str(x) != "nan"]
        self.get_body_part_names()
        self.get_bp_headers()
        self.bp_col_names = self.x_cols + self.y_cols + self.p_cols
        self.pose_setting = self.read_config_entry(
            config=self.config,
            section=ConfigKey.CREATE_ENSEMBLE_SETTINGS.value,
            option=ConfigKey.POSE_SETTING.value,
            data_type=Dtypes.STR.value,
        )
        self.roi_coordinates_path = os.path.join(
            self.logs_path, Paths.ROI_DEFINITIONS.value
        )
        self.video_dir = os.path.join(self.project_path, "videos")
        self.clf_validation_dir = os.path.join(
            self.project_path, Paths.CLF_VALIDATION_DIR.value
        )
        self.clf_data_validation_dir = os.path.join(
            self.project_path, "csv", "validation"
        )
        self.annotated_frm_dir = os.path.join(
            self.project_path, Paths.ANNOTATED_FRAMES_DIR.value
        )
        self.single_validation_video_save_dir = os.path.join(
            self.project_path, Paths.SINGLE_CLF_VALIDATION.value
        )
        self.data_table_path = os.path.join(self.project_path, Paths.DATA_TABLE.value)
        self.check_multi_animal_status()
        self.multiprocess_chunksize = Defaults.CHUNK_SIZE.value
        self.maxtasksperchild = Defaults.MAX_TASK_PER_CHILD.value
        self.clr_lst = create_color_palettes(self.animal_cnt, int(len(self.x_cols)) + 1)
        self.animal_bp_dict = self.create_body_part_dictionary(
            self.multi_animal_status,
            self.multi_animal_id_list,
            self.animal_cnt,
            self.x_cols,
            self.y_cols,
            self.p_cols,
            self.clr_lst,
        )
        self.project_bps = list(set([x[:-2] for x in self.bp_headers]))
        self.color_dict = get_color_dict()
        if create_logger:
            self.create_logger()
        self.emojis = get_emojis()
        if read_video_info:
            self.video_info_df = self.read_video_info_csv(
                file_path=self.video_info_path
            )

    def read_roi_data(self) -> None:
        """
        Method to read in ROI definitions from SimBA project
        """

        if not os.path.isfile(self.roi_coordinates_path):
            raise NoROIDataError(
                msg="SIMBA ERROR: No ROI definitions were found in your SimBA project. Please draw some ROIs before analyzing your ROI data",
                source=self.__class__.__name__,
            )
        else:

            self.rectangles_df = pd.read_hdf(
                self.roi_coordinates_path, key=Keys.ROI_RECTANGLES.value
            )
            if ("Center_X" in self.rectangles_df.columns) and (
                self.rectangles_df["Center_X"].isnull().values.any()
            ):
                for idx, row in self.rectangles_df.iterrows():
                    self.rectangles_df.loc[idx]["Center_X"] = row["Tags"]["Center tag"][
                        0
                    ]
                    self.rectangles_df.loc[idx]["Center_Y"] = row["Tags"]["Center tag"][
                        1
                    ]
            self.circles_df = pd.read_hdf(
                self.roi_coordinates_path, key=Keys.ROI_CIRCLES.value
            ).dropna(how="any")
            self.polygon_df = pd.read_hdf(
                self.roi_coordinates_path, key=Keys.ROI_POLYGONS.value
            )
            if "Center_XCenter_Y" in self.polygon_df.columns:
                self.polygon_df = self.polygon_df.drop(["Center_XCenter_Y"], axis=1)
            self.polygon_df = self.polygon_df.dropna(how="any")
            self.shape_names = list(
                itertools.chain(
                    self.rectangles_df["Name"].unique(),
                    self.circles_df["Name"].unique(),
                    self.polygon_df["Name"].unique(),
                )
            )
            self.roi_dict = {
                Keys.ROI_RECTANGLES.value: self.rectangles_df,
                Keys.ROI_CIRCLES.value: self.circles_df,
                Keys.ROI_POLYGONS.value: self.polygon_df,
            }
            self.roi_types_names_lst = set()
            for idx, r in self.roi_dict[Keys.ROI_RECTANGLES.value].iterrows():
                self.roi_types_names_lst.add(f'Rectangle: {r["Name"]}')
            for idx, r in self.roi_dict[Keys.ROI_CIRCLES.value].iterrows():
                self.roi_types_names_lst.add(f'Circle: {r["Name"]}')
            for idx, r in self.roi_dict[Keys.ROI_POLYGONS.value].iterrows():
                self.roi_types_names_lst.add(f'Polygon: {r["Name"]}')
            self.roi_types_names_lst = list(self.roi_types_names_lst)
            for shape_type, shape_data in self.roi_dict.items():
                if shape_type == Keys.ROI_CIRCLES.value:
                    self.roi_dict[Keys.ROI_CIRCLES.value]["Center_X"] = self.roi_dict[
                        Keys.ROI_CIRCLES.value
                    ]["centerX"]
                    self.roi_dict[Keys.ROI_CIRCLES.value]["Center_Y"] = self.roi_dict[
                        Keys.ROI_CIRCLES.value
                    ]["centerY"]
                elif shape_type == Keys.ROI_RECTANGLES.value:
                    self.roi_dict[Keys.ROI_RECTANGLES.value][
                        "Center_X"
                    ] = self.roi_dict[Keys.ROI_RECTANGLES.value]["Bottom_right_X"] - (
                        (
                            self.roi_dict[Keys.ROI_RECTANGLES.value]["Bottom_right_X"]
                            - self.roi_dict[Keys.ROI_RECTANGLES.value]["width"]
                        )
                        / 2
                    )
                    self.roi_dict[Keys.ROI_RECTANGLES.value][
                        "Center_Y"
                    ] = self.roi_dict[Keys.ROI_RECTANGLES.value]["Bottom_right_Y"] - (
                        (
                            self.roi_dict[Keys.ROI_RECTANGLES.value]["Bottom_right_Y"]
                            - self.roi_dict[Keys.ROI_RECTANGLES.value]["height"]
                        )
                        / 2
                    )
                elif shape_type == Keys.ROI_POLYGONS.value:
                    try:
                        self.roi_dict[Keys.ROI_POLYGONS.value]["Center_X"] = (
                            self.roi_dict[Keys.ROI_POLYGONS.value]["Center_X"]
                        )
                        self.roi_dict[Keys.ROI_POLYGONS.value]["Center_Y"] = (
                            self.roi_dict[Keys.ROI_POLYGONS.value]["Center_Y"]
                        )
                    except KeyError:
                        pass
            self.video_names_w_rois = set(
                list(self.rectangles_df["Video"])
                + list(self.circles_df["Video"])
                + list(self.polygon_df["Video"])
            )

    def get_all_clf_names(self) -> List[str]:
        """
        Helper to return all classifier names in SimBA project

        :return List[str]:
        """

        model_names = []
        for i in range(self.clf_cnt):
            entry_name = "target_name_{}".format(str(i + 1))
            model_names.append(
                self.read_config_entry(
                    self.config,
                    ConfigKey.SML_SETTINGS.value,
                    entry_name,
                    data_type=Dtypes.STR.value,
                )
            )
        return model_names

    def insert_column_headers_for_outlier_correction(
        self, data_df: pd.DataFrame, new_headers: List[str], filepath: str
    ) -> pd.DataFrame:
        """
        Helper to insert new column headers onto a dataframe.

        :param pd.DataFrame data_df:  Dataframe where headers to to-bo replaced.
        :param List[str] new_headers:  Names of new headers.
        :param str filepath:  Path to where ``data_df`` is stored on disk
        :returns pd.DataFrame: Dataframe with new headers
        :raises DataHeaderWarning: If new headers are fewer/more than columns in dataframe

        :example:
        >>> df = pd.DataFrame(data=[[1, 2, 3], [1, 2, 3]], columns=['Feature_1', 'Feature_2', 'Feature_3'])
        >>> ConfigReader.insert_column_headers_for_outlier_correction(data_df=df, new_headers=['Feature_4', 'Feature_5', 'Feature_6'], filepath='test/my_test_file.csv')
        """

        if len(new_headers) != len(data_df.columns):
            difference = int(len(data_df.columns) - len(new_headers))
            bp_missing = int(abs(difference) / 3)
            if difference < 0:
                raise DataHeaderError(
                    msg=f"SIMBA ERROR: SimBA expects {len(new_headers)} columns of data inside the files within project_folder/csv/input_csv directory. However, within file {filepath} file, SimBA found {len(data_df.columns)} columns. Thus, there is {abs(difference)} missing data columns in the imported data, which may represent {int(bp_missing)} bodyparts if each body-part has an x, y and p value. Either revise the SimBA project pose-configuration with {bp_missing} less body-part, or include {bp_missing} more body-part in the imported data",
                    source=self.__class__.__name__,
                )
            else:
                raise DataHeaderError(
                    msg=f"SIMBA ERROR: SimBA expects {len(new_headers)} columns of data inside the files within project_folder/csv/input_csv directory. However, within file {filepath} file, SimBA found {len(data_df.columns)} columns. Thus, there is {abs(difference)} more data columns in the imported data than anticipated, which may represent {int(bp_missing)} bodyparts if each body-part has an x, y and p value. Either revise the SimBA project pose-configuration with {bp_missing} more body-part, or include {bp_missing} less body-part in the imported data",
                    source=self.__class__.__name__,
                )
        else:
            data_df.columns = new_headers
            return data_df

    def get_number_of_header_columns_in_df(self, df: pd.DataFrame) -> int:
        """
        Helper to find the count of non-numerical rows at the top of a dataframe.

        :param pd.DataFrame data_df: Dataframe to find the count non-numerical header rows in.
        :returns int
        :raises DataHeaderError: All rows are non-numerical.

        :example:
        >>> ConfigReader.get_number_of_header_columns_in_df(df=pd.DataFrame(data=[[1, 2, 3], [1, 2, 3]]))
        >>> 0
        >>> ConfigReader.get_number_of_header_columns_in_df(df=pd.DataFrame(data=[['Head_1', 'Body_2', 'Tail_3'], ['Some_nonsense', 'A_mistake', 'Maybe_multi_headers?'], [11, 99, 109], [122, 43, 2091]]))
        >>> 2
        """

        for i in range(len(df)):
            try:
                _ = df.iloc[i:].apply(pd.to_numeric).reset_index(drop=True)
                return i
            except ValueError:
                pass
        raise DataHeaderError(
            msg="Could find the count of header columns in dataframe. All values appear to be non-numeric",
            source=self.__class__.__name__,
        )

    def find_video_of_file(
        self,
        video_dir: Union[str, os.PathLike],
        filename: str,
        raise_error: bool = False,
    ) -> Union[str, os.PathLike]:
        """
        Helper to find the video file representing a known data file basename.

        :param Union[str, os.PathLike] video_dir: Directory holding putative video file.
        :param str filename: Data file name, e.g., ``Video_1``.
        :param bool raise_error: If True, raise error if no video can be found.
        :return Union[str, os.PathLike]: Path to video file.


        :example:
        >>> config_reader = ConfigReader(config_path='My_SimBA_Config')
        >>> config_reader.find_video_of_file(video_dir=config_reader.video_dir, filename='Video1')
        >>> '/project_folder/videos/Video1.mp4'
        """
        try:
            all_files_in_video_folder = [
                f for f in next(os.walk(video_dir))[2] if not f[0] == "."
            ]
        except StopIteration:
            raise NoFilesFoundError(
                msg=f"No files found in the {video_dir} directory",
                source=self.__class__.__name__,
            )
        all_files_in_video_folder = [
            os.path.join(video_dir, x) for x in all_files_in_video_folder
        ]
        return_path = None
        for file_path in all_files_in_video_folder:
            _, video_filename, ext = get_fn_ext(file_path)
            if (video_filename == filename) and (
                (ext.lower() == ".mp4") or (ext.lower() == ".avi")
            ):
                return_path = file_path

        if return_path is None and not raise_error:
            NoFileFoundWarning(
                msg=f"SimBA could not find a video file representing {filename} in the project video directory",
                source=self.__class__.__name__,
            )
        if return_path is None and raise_error:
            raise NoFilesFoundError(
                msg=f"SimBA could not find a video file representing {filename} in the project video directory",
                source=self.__class__.__name__,
            )
        return return_path

    def add_missing_ROI_cols(self, shape_df: pd.DataFrame) -> pd.DataFrame:
        """
        Helper to add missing ROI definitions (``Color BGR``, ``Thickness``, ``Color name``) in ROI info
        dataframes created by the first version of the SimBA ROI user-interface but analyzed using newer versions of SimBA.


        :param pd.DataFrame shape_df: Dataframe holding ROI definitions.
        :return pd.DataFrame with ``Color BGR``, ``Thickness``, ``Color name`` fields
        """

        if not "Color BGR" in shape_df.columns:
            shape_df["Color BGR"] = [(255, 255, 255)] * len(shape_df)
        if not "Thickness" in shape_df.columns:
            shape_df["Thickness"] = [5] * len(shape_df)
        if not "Color name" in shape_df.columns:
            shape_df["Color name"] = "White"

        return shape_df

    def remove_a_folder(
        self, folder_dir: str, raise_error: Optional[bool] = False
    ) -> None:
        """
        Helper to remove single directory.

        :param folder_dir: Directory to remove.
        :param bool raise_error: If True, raise ``NotDirectoryError`` error of folder does not exist.
        :raises NotDirectoryError: If ``raise_error`` and directory does not exist.

        :example:
        >>> self.remove_a_folder(folder_dir'gerbil/gerbil_data/featurized_data/temp')
        """
        shutil.rmtree(folder_dir, ignore_errors=True)

    def remove_multiple_folders(
        self, folders: List[os.PathLike], raise_error: Optional[bool] = False
    ) -> None:
        """
        Helper to remove multiple directories.

        :param folders List[os.PathLike]: List of directory paths.
        :param bool raise_error: If True, raise ``NotDirectoryError`` error of folder does not exist. if False, then pass. Default False.
        :raises NotDirectoryError: If ``raise_error`` and directory does not exist.

        :example:
        >>> self.remove_multiple_folders(folders= ['gerbil/gerbil_data/featurized_data/temp'])
        """
        folders = [x for x in folders if x is not None]
        for folder_path in folders:
            if raise_error and not os.path.isdir(folder_path):
                raise NotDirectoryError(
                    msg=f"Cannot delete directory {folder_path}. The directory does not exist.",
                    source=self.__class__.__name__,
                )
            if os.path.isdir(folder_path):
                shutil.rmtree(folder_path, ignore_errors=True)
            else:
                pass

    def find_animal_name_from_body_part_name(self, bp_name: str, bp_dict: dict) -> str:
        """
        Given body-part name and animal body-part dict, returns the animal name

        :param str bp_name: Name of the body-part. E.g., ``Ear_1``.
        :param dict bp_dict: Nested dict holding animal names as keys and body-part names as and coordinates as values. Created by :meth:`simba.mixins.config_reader.ConfigReader.create_body_part_dictionary`
        :returns str

        :example:
        >>> config_reader = ConfigReader(config_path='tests/data/test_projects/two_c57/project_folder/project_config.ini')
        >>> ConfigReader.find_animal_name_from_body_part_name(bp_name='Ear_1', bp_dict=config_reader.animal_bp_dict)
        >>> 'simon'
        """

        for animal_name, animal_bps in bp_dict.items():
            if bp_name in [x[:-2] for x in animal_bps["X_bps"]]:
                return animal_name

    def create_body_part_dictionary(
        self,
        multi_animal_status: bool,
        animal_id_lst: list,
        animal_cnt: int,
        x_cols: List[str],
        y_cols: List[str],
        p_cols: Optional[List[str]] = None,
        colors: Optional[List[List[Tuple[int, int, int]]]] = None,
    ) -> Dict[str, Union[List[str], List[Tuple]]]:
        """
        Helper to create dict of dict lookup of body-parts where the keys are animal names, and
        values are the body-part names.

        :param bool multi_animal_status: If True, it is a multi-animal SimBA project.
        :param List[str] multi_animal_id_lst: Animal names. Eg., ['Simon, 'JJ']. Note: If a single animal project, this will be overridden and set to `Animal_1`.
        :param int animal_cnt: Number of animals in the SimBA project.
        :param List[str] x_cols: column names for body-part coordinates on x-axis. Returned by  :meth:`simba.mixins.config_reader.ConfigReader.get_body_part_names`
        :param List[str] y_cols: column names for body-part coordinates on y-axis. Returned by  :meth:`simba.mixins.config_reader.ConfigReader.get_body_part_names`
        :param List[str] p_cols: column names for body-part pose-estimation probability values. Returned by  :meth:`simba.mixins.config_reader.ConfigReader.get_body_part_names`
        :param Optional[List[List[Tuple[int, int, int]]]] colors: Optional bgr colors to associate with the body-parts. Returned by :meth:`simba.utils.data.create_color_palettes`.
        :returns dict

        :example:
        >>> ConfigReader.create_body_part_dictionary(multi_animal_status=True, animal_id_lst=['simon',])
        >>> {'simon': {'X_bps': ['Nose_1_x', 'Ear_left_1_x', 'Ear_right_1_x', 'Center_1_x', 'Lat_left_1_x', 'Lat_right_1_x', 'Tail_base_1_x', 'Tail_end_1_x'], 'Y_bps': ['Nose_1_y', 'Ear_left_1_y', 'Ear_right_1_y', 'Center_1_y', 'Lat_left_1_y', 'Lat_right_1_y', 'Tail_base_1_y', 'Tail_end_1_y'], 'colors': [[255.0, 0.0, 255.0], [223.125, 31.875, 255.0], [191.25, 63.75, 255.0], [159.375, 95.625, 255.0], [127.5, 127.5, 255.0], [95.625, 159.375, 255.0], [63.75, 191.25, 255.0], [31.875, 223.125, 255.0], [0.0, 255.0, 255.0]], 'P_bps': ['Nose_1_p', 'Ear_left_1_p', 'Ear_right_1_p', 'Center_1_p', 'Lat_left_1_p', 'Lat_right_1_p', 'Tail_base_1_p', 'Tail_end_1_p']}, 'jj': {'X_bps': ['Nose_2_x', 'Ear_left_2_x', 'Ear_right_2_x', 'Center_2_x', 'Lat_left_2_x', 'Lat_right_2_x', 'Tail_base_2_x', 'Tail_end_2_x'], 'Y_bps': ['Nose_2_y', 'Ear_left_2_y', 'Ear_right_2_y', 'Center_2_y', 'Lat_left_2_y', 'Lat_right_2_y', 'Tail_base_2_y', 'Tail_end_2_y'], 'colors': [[102.0, 127.5, 0.0], [102.0, 143.4375, 31.875], [102.0, 159.375, 63.75], [102.0, 175.3125, 95.625], [102.0, 191.25, 127.5], [102.0, 207.1875, 159.375], [102.0, 223.125, 191.25], [102.0, 239.0625, 223.125], [102.0, 255.0, 255.0]], 'P_bps': ['Nose_2_p', 'Ear_left_2_p', 'Ear_right_2_p', 'Center_2_p', 'Lat_left_2_p', 'Lat_right_2_p', 'Tail_base_2_p', 'Tail_end_2_p']}}
        """

        animal_bp_dict = {}

        if multi_animal_status:
            for animal in range(animal_cnt):
                animal_bp_dict[animal_id_lst[animal]] = {}
                animal_bp_dict[animal_id_lst[animal]]["X_bps"] = [
                    i for i in x_cols if animal_id_lst[animal] in i
                ]
                animal_bp_dict[animal_id_lst[animal]]["Y_bps"] = [
                    i for i in y_cols if animal_id_lst[animal] in i
                ]
                if colors:
                    animal_bp_dict[animal_id_lst[animal]]["colors"] = colors[animal]
                if p_cols:
                    animal_bp_dict[animal_id_lst[animal]]["P_bps"] = [
                        i for i in p_cols if animal_id_lst[animal] in i
                    ]
                if not animal_bp_dict[animal_id_lst[animal]]["X_bps"]:
                    multi_animal_status = False
                    break

        if not multi_animal_status:
            if animal_cnt > 1:
                for animal in range(animal_cnt):
                    currAnimalName = f"Animal_{str(animal + 1)}"
                    search_string_x = f"_{str(animal + 1)}_x"
                    search_string_y = f"_{str(animal + 1)}_y"
                    search_string_p = f"_{str(animal + 1)}_p"
                    animal_bp_dict[currAnimalName] = {}
                    animal_bp_dict[currAnimalName]["X_bps"] = [
                        i for i in x_cols if i.endswith(search_string_x)
                    ]
                    animal_bp_dict[currAnimalName]["Y_bps"] = [
                        i for i in y_cols if i.endswith(search_string_y)
                    ]
                    if colors:
                        animal_bp_dict[currAnimalName]["colors"] = colors[animal]
                    if p_cols:
                        animal_bp_dict[currAnimalName]["P_bps"] = [
                            i for i in p_cols if i.endswith(search_string_p)
                        ]
                if animal_id_lst[0] != "":
                    for animal in range(len(animal_id_lst)):
                        currAnimalName = f"Animal_{str(animal + 1)}"
                        animal_bp_dict[animal_id_lst[animal]] = animal_bp_dict.pop(
                            currAnimalName
                        )

            else:
                animal_bp_dict["Animal_1"] = {}
                animal_bp_dict["Animal_1"]["X_bps"] = [i for i in x_cols]
                animal_bp_dict["Animal_1"]["Y_bps"] = [i for i in y_cols]
                if colors:
                    animal_bp_dict["Animal_1"]["colors"] = colors[0]
                if p_cols:
                    animal_bp_dict["Animal_1"]["P_bps"] = [i for i in p_cols]
        return animal_bp_dict

    def get_body_part_names(self):
        """
        Helper to extract pose-estimation data field names (x, y, p)

        :example:
        >>> config_reader = ConfigReader(config_path='test/project_config.csv')
        >>> config_reader.get_body_part_names()
        """

        self.x_cols, self.y_cols, self.p_cols = [], [], []
        for bp in self.body_parts_lst:
            self.x_cols.append(f"{bp}_x")
            self.y_cols.append(f"{bp}_y")
            self.p_cols.append(f"{bp}_p")

    def drop_bp_cords(
        self, df: pd.DataFrame, raise_error: bool = False
    ) -> pd.DataFrame:
        """
        Helper to remove pose-estimation fields from dataframe.

        :param pd.DataFrame df: pandas dataframe containing pose-estimation fields (body-part x, y, p fields)
        :param bool raise_error: If True, raise error if body-parts cant be found. Else, print warning
        :return pd.DataFrame: ``df`` without pose-estimation fields

        :example:
        >>> config_reader = ConfigReader(config_path='test/project_folder/project_config.csv')
        >>> df = read_df(config_reader.machine_results_paths[0], file_type='csv')
        >>> df = config_reader.drop_bp_cords(df=df)
        """
        missing_body_part_fields = list(set(self.bp_col_names) - set(list(df.columns)))
        if len(missing_body_part_fields) > 0 and not raise_error:
            BodypartColumnNotFoundWarning(
                msg=f"SimBA could not drop body-part coordinates, some body-part names are missing in dataframe. SimBA expected the following body-parts, that could not be found inside the file: {missing_body_part_fields}",
                source=self.__class__.__name__,
            )
            return df.drop(self.bp_col_names, axis=1, errors="ignore")
        elif len(missing_body_part_fields) > 0 and raise_error:
            raise BodypartColumnNotFoundError(
                msg=f"SimBA could not drop body-part coordinates, some body-part names are missing in dataframe. SimBA expected the following body-parts, that could not be found inside the file: {missing_body_part_fields}",
                source=self.__class__.__name__,
            )
        else:
            return df.drop(self.bp_col_names, axis=1)

    def get_bp_headers(self) -> None:
        """
        Helper to create ordered list of all column header fields for SimBA project dataframes.

        >>> config_reader = ConfigReader(config_path='test/project_folder/project_config.ini')
        >>> config_reader.get_bp_headers()
        """

        self.bp_headers = []
        for bp in self.body_parts_lst:
            c1, c2, c3 = (f"{bp}_x", f"{bp}_y", f"{bp}_p")
            self.bp_headers.extend((c1, c2, c3))

    def read_config_entry(
        self,
        config: ConfigParser,
        section: str,
        option: str,
        data_type: Literal["str", "int", "float", "folder_path"],
        default_value: Optional[Any] = None,
        options: Optional[List[Any]] = None,
    ) -> Union[str, int, float]:
        """
        Helper to read entry from a configparser.ConfigParser object

        :param ConfigParser config: Parsed SimBA project config
        :param str section: Project config section name
        :param str option: Project config option name
        :param str data_type: Type of data. E.g., ``str``, ``int``, ``float``, ``folder_path``.
        :param Optional[Any] default_value: If entry cannot be found, then default to this value.
        :param Optional[List[Any]] options: If not ``None``, then list of viable entries.
        :raise InvalidInputError: If returned value is not in ``options``.
        :raise MissingProjectConfigEntryError: If no entry is found and no ``default_value`` is provided.

        :return Union[str, float, int, os.Pathlike]

        :example:
        >>> config = ConfigReader(config_path='tests/data/test_projects/two_c57/project_folder/project_config.ini')
        >>> config.read_config_entry(config=self.config, section='Multi animal IDs', option='id_list', data_type='str')
        >>> 'simon,jj'
        """
        try:
            if config.has_option(section, option):
                if data_type == Dtypes.FLOAT.value:
                    value = config.getfloat(section, option)
                elif data_type == Dtypes.INT.value:
                    value = config.getint(section, option)
                elif data_type == Dtypes.STR.value:
                    value = config.get(section, option).strip()
                elif data_type == Dtypes.FOLDER.value:
                    value = config.get(section, option).strip()
                    if not os.path.isdir(value):
                        raise NotDirectoryError(
                            msg=f"The SimBA config file includes paths to a folder ({value}) that does not exist.",
                            source=self.__class__.__name__,
                        )
                if options != None:
                    if value not in options:
                        raise InvalidInputError(
                            msg=f"{option} is set to {str(value)} in SimBA, but this is not among the valid options: ({options})",
                            source=self.__class__.__name__,
                        )
                    else:
                        return value
                return value
            elif default_value != None:
                return default_value
            else:
                raise MissingProjectConfigEntryError(
                    msg=f"SimBA could not find an entry for option {option} under section {section} in the project_config.ini. Please specify the settings in the settings menu.",
                    source=self.__class__.__name__,
                )
        except ValueError:
            if default_value != None:
                return default_value
            else:
                raise MissingProjectConfigEntryError(
                    msg=f"SimBA could not find an entry for option {option} under section {section} in the project_config.ini. Please specify the settings in the settings menu.",
                    source=self.__class__.__name__,
                )

    def read_video_info_csv(self, file_path: str) -> pd.DataFrame:
        """
        Helper to read the project_folder/logs/video_info.csv of the SimBA project in as a pd.DataFrame
        Parameters
        ----------
        file_path: str

        Returns
        -------
        pd.DataFrame
        """

        if not os.path.isfile(file_path):
            raise NoFilesFoundError(
                msg=f"Could not find the video_info.csv table in your SimBA project. Create it using the [Video parameters] tab. SimBA expects the file at location {file_path}",
                source=self.__class__.__name__,
            )
        info_df = pd.read_csv(file_path)
        for c in [
            "Video",
            "fps",
            "Resolution_width",
            "Resolution_height",
            "Distance_in_mm",
            "pixels/mm",
        ]:
            if c not in info_df.columns:
                raise ParametersFileError(
                    msg=f'The project "project_folder/logs/video_info.csv" does not not have an anticipated header ({c}). Please re-create the file and make sure each video has a {c} value',
                    source=self.__class__.__name__,
                )
        info_df["Video"] = info_df["Video"].astype(str)
        for c in [
            "fps",
            "Resolution_width",
            "Resolution_height",
            "Distance_in_mm",
            "pixels/mm",
        ]:
            try:
                info_df[c] = info_df[c].astype(float)
            except:
                raise ParametersFileError(
                    msg=f'One or more values in the {c} column of the "project_folder/logs/video_info.csv" file could not be interpreted as a numeric value. Please re-create the file and make sure the entries in the {c} column are all numeric.',
                    source=self.__class__.__name__,
                )
        if info_df["fps"].min() <= 1:
            InvalidValueWarning(
                msg="Videos in your SimBA project have an FPS of 1 or less. Please use videos with more than one frame per second, or correct the inaccurate fps inside the `project_folder/logs/videos_info.csv` file",
                source=self.__class__.__name__,
            )
        return info_df

    def read_video_info(
        self, video_name: str, raise_error: Optional[bool] = True
    ) -> (pd.DataFrame, float, float):
        """
        Helper to read the meta-data (pixels per mm, resolution, fps) from the video_info.csv for a single input file.

        :param str video_name: The name of the video without extension to get the metadata for
        :param Optional[bool] raise_error: If True, raise error if video info for the video name cannot be found. Default: True.
        :raise ParametersFileError: If ``raise_error`` and video metadata info is not found
        :raise DuplicationError: If file contains multiple entries for the same video.
        :return (pd.DataFrame, float, float) representing all video info, pixels per mm, and fps
        """

        video_settings = self.video_info_df.loc[
            self.video_info_df["Video"] == video_name
        ]
        if len(video_settings) > 1:
            raise DuplicationError(
                msg=f"SimBA found multiple rows in the project_folder/logs/video_info.csv named {str(video_name)}. Please make sure that each video name is represented ONCE in the video_info.csv",
                source=self.__class__.__name__,
            )
        elif len(video_settings) < 1:
            if raise_error:
                raise ParametersFileError(
                    msg=f"SimBA could not find {str(video_name)} in the video_info.csv file. Make sure all videos analyzed are represented in the project_folder/logs/video_info.csv file.",
                    source=self.__class__.__name__,
                )
            else:
                return (None, None, None)
        else:
            try:
                px_per_mm = float(video_settings["pixels/mm"])
                fps = float(video_settings["fps"])
                return video_settings, px_per_mm, fps
            except TypeError:
                raise ParametersFileError(
                    msg=f"Make sure the videos that are going to be analyzed are represented with APPROPRIATE VALUES inside the project_folder/logs/video_info.csv file in your SimBA project. Could not interpret the fps, pixels per millimeter and/or fps as numerical values for video {video_name}",
                    source=self.__class__.__name__,
                )

    def check_multi_animal_status(self) -> None:
        """
        Helper to check if the project is a multi-animal SimBA project.

        """
        multi_animal_id_lst = []
        if not self.config.has_section("Multi animal IDs"):
            for animal in range(self.animal_cnt):
                multi_animal_id_lst.append("Animal_" + str(animal + 1))
            multi_animal_status = False

        else:
            multi_animal_id_str = self.read_config_entry(
                config=self.config,
                section="Multi animal IDs",
                option="id_list",
                data_type="str",
            )
            multi_animal_id_lst = [x.lstrip() for x in multi_animal_id_str.split(",")]
            multi_animal_id_lst = [x for x in multi_animal_id_lst if x != "None"]
            if (self.animal_cnt > 1) and (len(multi_animal_id_lst) > 1):
                multi_animal_status = True
            else:
                for animal in range(self.animal_cnt):
                    multi_animal_id_lst.append("Animal_{}".format(str(animal + 1)))
                multi_animal_status = False

        self.multi_animal_status = multi_animal_status
        self.multi_animal_id_list = multi_animal_id_lst[: self.animal_cnt]

    def remove_roi_features(self, data_dir: Union[str, os.PathLike]) -> None:
        """
        Helper to remove ROI-based features from datasets within a directory. The identified ROI-based fields are move to the
        ``project_folder/logs/ROI_data_{datetime}`` directory.

        .. note::
           ROI-based features are identified based on the combined criteria of (i) The prefix of the field is a named ROI in
           the ``project_folder/logs/ROI_definitions.h5`` file, and (ii) the suffix of the field is contained in the ['in zone',
           'n zone_cumulative_time', 'in zone_cumulative_percent', 'distance', 'facing']

        :param Union[str, os.PathLike] data_dir: directory with data to remove ROi features from.

        :example:
        >>> self.remove_roi_features('/project_folder/csv/features_extracted')

        """
        ROI_COL_SUFFIXES = [
            "in zone",
            "n zone_cumulative_time",
            "in zone_cumulative_percent",
            "distance",
            "facing",
        ]

        check_if_dir_exists(in_dir=data_dir)
        timer = SimbaTimer(start=True)
        filepaths = glob.glob(data_dir + f"/*.{self.file_type}")
        roi_dir = os.path.join(
            self.project_path,
            "logs",
            f'ROI_data_{datetime.now().strftime("%Y%m%d%H%M%S")}',
        )
        check_if_filepath_list_is_empty(
            filepaths=filepaths,
            error_msg=f"No .{self.file_type} files found in {data_dir}.",
        )
        for file_cnt, file_path in enumerate(filepaths):
            _, video_name, _ = get_fn_ext(filepath=file_path)
            roi_cols = set()
            df = read_df(file_path=file_path, file_type=self.file_type)
            for shape_name in self.shape_names:
                roi_cols.update([x for x in df.columns if shape_name in x])
            for suffix in ROI_COL_SUFFIXES:
                roi_cols.update([x for x in roi_cols if x.endswith(suffix)])
            if len(roi_cols) == 0:
                NoDataFoundWarning(
                    msg=f"NO ROI data found in {file_path}",
                    source=self.__class__.__name__,
                )
            else:
                roi_df = df[list(roi_cols)]
                df = df.drop(list(roi_cols), axis=1)
                roi_save_path = os.path.join(roi_dir, video_name + "." + self.file_type)
                if not os.path.exists(roi_dir):
                    os.makedirs(roi_dir)
                write_df(df=roi_df, file_type=self.file_type, save_path=roi_save_path)
                write_df(df=df, file_type=self.file_type, save_path=file_path)
                print(f"ROI features (N={len(roi_cols)}) removed from {video_name}...")
        timer.stop_timer()
        stdout_success(
            msg=f"ROI features removed from {len(filepaths)} files in the {data_dir} directory. The ROI features has been moved to the {roi_dir} directory",
            elapsed_time=timer.elapsed_time_str,
            source=self.__class__.__name__,
        )

    def create_log_msg_from_init_args(self, locals: dict):

        def has_methods_or_classes(obj):
            return any(callable(attr) for attr in dir(obj))

        def remove_keys_with_methods_or_classes(dictionary):
            return {
                key: value
                for key, value in dictionary.items()
                if not has_methods_or_classes(value)
            }

        msg = ""
        locals.pop("self", None)
        locals.pop("__class__", None)
        locals = remove_keys_with_methods_or_classes(locals)

        for cnt, (k, v) in enumerate(locals.items()):
            msg += f"{k}: {v}"
            if cnt != len(list(locals.keys())) - 1:
                msg += ", "
        return msg

    def create_logger(self) -> None:
        if self.__class__ not in logging.Logger.manager.loggerDict.keys():
            log_config = get_log_config()
            log_config["handlers"]["file_handler"]["filename"] = os.path.join(
                self.logs_path, "project_log.log"
            )
            logging.config.dictConfig(log_config)
            self.logger = logging.getLogger(str(self.__class__))
        if not os.path.isfile(os.path.join(self.logs_path, "project_log.log")):
            self.logger.info(f"Logging initiated in project {self.project_path}...")

    def create_third_part_append_logger(self, path: str) -> None:
        logger = logging.getLogger()
        logger.setLevel(logging.INFO)
        handler = logging.FileHandler(filename=path, encoding="utf-8")  # or whatever
        handler.setFormatter(logging.Formatter("%(asctime)s %(levelname)s %(message)s"))
        logger.addHandler(handler)

    def add_video_to_video_info(
        self,
        video_name: str,
        fps: float,
        width: int,
        height: int,
        pixels_per_mm: int,
        distance_mm: float = 0.0,
    ) -> None:
        self.video_info_df.loc[len(self.video_info_df)] = [
            video_name,
            fps,
            width,
            height,
            distance_mm,
            pixels_per_mm,
        ]
        self.video_info_df.drop_duplicates(subset=["Video"], inplace=True, keep="last")
        self.video_info_df = self.video_info_df.set_index("Video")
        try:
            self.video_info_df.to_csv(
                os.path.join(self.project_path, Paths.VIDEO_INFO.value)
            )
        except:
            raise PermissionError(
                msg="SimBA tried to write to project_folder/logs/video_info.csv, but was not allowed. If this file is open in another program, try closing it.",
                source=self.__class__.__name__,
            )
        stdout_success(
            msg="Video info has been UPDATED at project_folder/logs/video_info.csv",
            source=self.__class__.__name__,
        )


# config = ConfigReader(config_path='/Users/simon/Desktop/envs/simba_dev/tests/data/test_projects/two_c57/project_folder/project_config.ini', read_video_info=False)

# config.read_roi_data()


# config = ConfigReader(config_path='/Users/simon/Desktop/envs/troubleshooting/Nastacia_unsupervised/project_folder/project_config.ini', read_video_info=False)
# config.read_roi_data()
# config.remove_roi_features('/Users/simon/Desktop/envs/troubleshooting/Nastacia_unsupervised/project_folder/csv/features_extracted')

# remove_roi_features(
#     config_path='/Users/simon/Desktop/envs/troubleshooting/Nastacia_unsupervised/project_folder/project_config.ini',
#     data_dir='/Users/simon/Desktop/envs/troubleshooting/Nastacia_unsupervised/project_folder/csv/features_extracted')

# test = ConfigReader(config_path='/Users/simon/Desktop/envs/troubleshooting/two_black_animals_14bp/project_folder/project_config.ini')



__author__ = "Tzuk Polinsky"

import itertools
import os
from typing import Union

import numpy as np
import pandas as pd

from simba.mixins.config_reader import ConfigReader
from simba.mixins.feature_extraction_mixin import FeatureExtractionMixin
from simba.utils.checks import check_if_filepath_list_is_empty
from simba.utils.enums import ConfigKey, Dtypes
from simba.utils.errors import AnimalNumberError
from simba.utils.printing import SimbaTimer, stdout_success
from simba.utils.read_write import get_fn_ext, read_df


class DirectingAnimalsToBodyPartAnalyzer(ConfigReader, FeatureExtractionMixin):
    """
    Calculate when animals are directing towards their own body-parts. Results are stored in
    the ``project_folder/logs/directionality_dataframes`` directory of the SimBA project.

    :parameter str config_path: path to SimBA project config file in Configparser format

    .. important::
       Requires the pose-estimation data for the ``left ear``, ``right ear`` and ``nose`` of each individual animals.
       `Tutorial <https://github.com/sgoldenlab/simba/blob/master/docs/ROI_tutorial.md#part-3-generating-features-from-roi-data>`__.
       `Expected output <https://github.com/sgoldenlab/simba/blob/master/misc/Direction_data_example.csv>`__.

    Examples
    -----
    >>> directing_analyzer = DirectingAnimalsToBodyPartAnalyzer(config_path='MyProjectConfig')
    >>> directing_analyzer.process_directionality()
    >>> directing_analyzer.create_directionality_dfs()
    >>> directing_analyzer.save_directionality_dfs()
    >>> directing_analyzer.summary_statistics()
    """

    def __init__(self, config_path: Union[str, os.PathLike]):
        super().__init__(config_path=config_path)
        if not os.path.exists(self.directionality_df_dir):
            os.makedirs(self.directionality_df_dir)
        check_if_filepath_list_is_empty(
            filepaths=self.outlier_corrected_paths,
            error_msg=f"SIMBA ERROR: No data found in the {self.outlier_corrected_dir} directory",
        )

        print(f"Processing {str(len(self.outlier_corrected_paths))} video(s)...")

    def process_directionality(self):
        """
        Method to compute when animals are directing towards their own body-parts.

        Returns
        -------
        Attribute: dict
            results_dict
        """

        self.results_dict = {}
        bp_x_name = self.bodypart_direction + "_x"
        bp_y_name = self.bodypart_direction + "_y"
        for file_cnt, file_path in enumerate(self.outlier_corrected_paths):
            video_timer = SimbaTimer(start=True)
            _, video_name, _ = get_fn_ext(file_path)
            self.results_dict[video_name] = {}
            data_df = read_df(file_path, self.file_type)
            direct_bp_dict = self.check_directionality_cords()
            for key, animal in direct_bp_dict.items():
                result_key = "{} {} {}".format(
                    key,
                    "directing towards",
                    self.bodypart_direction,
                )
                self.results_dict[video_name][result_key] = {}
                ear_left_arr = data_df[
                    [
                        animal["Ear_left"]["X_bps"],
                        animal["Ear_left"]["Y_bps"],
                    ]
                ].to_numpy()
                ear_right_arr = data_df[
                    [
                        animal["Ear_right"]["X_bps"],
                        animal["Ear_right"]["Y_bps"],
                    ]
                ].to_numpy()
                nose_arr = data_df[
                    [
                        animal["Nose"]["X_bps"],
                        animal["Nose"]["Y_bps"],
                    ]
                ].to_numpy()
                target_cord_arr = data_df[[bp_x_name, bp_y_name]].to_numpy()
                direction_data = self.jitted_line_crosses_to_nonstatic_targets(
                    left_ear_array=ear_left_arr,
                    right_ear_array=ear_right_arr,
                    nose_array=nose_arr,
                    target_array=target_cord_arr,
                )
                x_min = np.minimum(direction_data[:, 1], nose_arr[:, 0])
                y_min = np.minimum(direction_data[:, 2], nose_arr[:, 1])
                delta_x = abs((direction_data[:, 1] - nose_arr[:, 0]) / 2)
                delta_y = abs((direction_data[:, 2] - nose_arr[:, 1]) / 2)
                x_middle, y_middle = np.add(x_min, delta_x), np.add(y_min, delta_y)
                direction_data = np.concatenate(
                    (y_middle.reshape(-1, 1), direction_data), axis=1
                )
                direction_data = np.concatenate(
                    (x_middle.reshape(-1, 1), direction_data), axis=1
                )
                direction_data = np.delete(direction_data, [2, 3, 4], 1)
                direction_data = np.hstack((direction_data, target_cord_arr))
                bp_data = pd.DataFrame(
                    direction_data,
                    columns=["Eye_x", "Eye_y", "Directing_BOOL", bp_x_name, bp_y_name],
                )
                bp_data = bp_data[
                    ["Eye_x", "Eye_y", bp_x_name, bp_y_name, "Directing_BOOL"]
                ]
                self.results_dict[video_name][result_key][bp_x_name[:-2]] = bp_data
            video_timer.stop_timer()
            print(
                "Direction analysis complete for video {} ({}/{}, elapsed time: {}s)...".format(
                    video_name,
                    str(file_cnt + 1),
                    str(len(self.outlier_corrected_paths)),
                    video_timer.elapsed_time_str,
                )
            )

    def create_directionality_dfs(self):
        """
        Method to transpose results created by :meth:`~simba.DirectingOtherAnimalsAnalyzer.process_directionality`.
        into dict of dataframes

        Returns
        -------
        Attribute: dict
            directionality_df_dict
        """

        print("Transposing body part directionality data...")
        self.directionality_df_dict = {}
        for video_name, video_data in self.results_dict.items():
            out_df_lst = []
            for animal_permutation, permutation_data in video_data.items():
                for bp_name, bp_data in permutation_data.items():
                    directing_df = bp_data.reset_index().rename(  # [bp_data["Directing_BOOL"] == 1]
                        columns={
                            "index": "Frame_#",
                            bp_name
                            + "_x": "Animal_{}_x".format(self.bodypart_direction),
                            bp_name
                            + "_y": "Animal_{}_y".format(self.bodypart_direction),
                        }
                    )
                    directing_df.insert(loc=0, column="Video", value=video_name)
                    out_df_lst.append(directing_df)
            self.directionality_df_dict[video_name] = pd.concat(out_df_lst, axis=0)
        stdout_success(msg="Transposing body part directionality data completed")

    def read_directionality_dfs(self):
        results = {}
        for file_cnt, file_path in enumerate(self.body_part_directionality_paths):
            video_timer = SimbaTimer(start=True)
            _, file_name, _ = get_fn_ext(file_path)
            results[file_name] = read_df(file_path, self.file_type)
            video_timer.stop_timer()
            print(
                "read body part directionality data completed for video {} ({}/{}, elapsed time: {}s)...".format(
                    file_name,
                    str(file_cnt + 1),
                    str(len(self.outlier_corrected_paths)),
                    video_timer.elapsed_time_str,
                )
            )
        stdout_success(msg="reading body part directionality data completed")
        return results

    def save_directionality_dfs(self):
        """
        Method to save result created by :meth:`~simba.DirectingOtherAnimalsAnalyzer.create_directionality_dfs`.
        into CSV files on disk. Results are stored in `project_folder/logs` directory of the SimBA project.

        Returns
        -------
        None
        """
        if not os.path.exists(self.body_part_directionality_df_dir):
            os.makedirs(self.body_part_directionality_df_dir)
        for video_name, video_data in self.directionality_df_dict.items():
            save_name = os.path.join(
                self.body_part_directionality_df_dir, video_name + ".csv"
            )
            video_data.to_csv(save_name)
            print(f"Detailed directional data saved for video {video_name}...")
        stdout_success(
            msg=f"All detailed directional data saved in the {self.body_part_directionality_df_dir} directory"
        )

    def summary_statistics(self):
        """
        Method to save aggregate statistics of data created by :meth:`~simba.DirectingOtherAnimalsAnalyzer.create_directionality_dfs`.
        into CSV files on disk. Results are stored in `project_folder/logs` directory of the SimBA project.

        Returns
        -------
        None
        """

        print("Computing summary statistics...")
        out_df_lst = []
        for video_name, video_data in self.results_dict.items():
            _, _, fps = self.read_video_info(video_name=video_name)
            for animal_permutation, permutation_data in video_data.items():
                idx_directing = set()
                for bp_name, bp_data in permutation_data.items():
                    idx_directing.update(
                        list(bp_data.index[bp_data["Directing_BOOL"] == 1])
                    )
                value = round(len(idx_directing) / fps, 3)
                out_df_lst.append(
                    pd.DataFrame(
                        [[video_name, animal_permutation, value]],
                        columns=["Video", "Animal permutation", "Value (s)"],
                    )
                )
        self.summary_df = (
            pd.concat(out_df_lst, axis=0)
            .sort_values(by=["Video", "Animal permutation"])
            .set_index("Video")
        )
        self.save_path = os.path.join(
            self.logs_path,
            "Body_part_directions_data_{}.csv".format(str(self.datetime)),
        )
        self.summary_df.to_csv(self.save_path)
        self.timer.stop_timer()
        stdout_success(
            msg=f"Summary body part directional statistics saved at {self.save_path}"
        )
        stdout_success(
            msg="All directional data saved in SimBA project",
            elapsed_time=self.timer.elapsed_time_str,
        )


# test = DirectingOtherAnimalsAnalyzer(config_path='/Users/simon/Desktop/envs/troubleshooting/two_black_animals_14bp/project_folder/project_config.ini')
# test.process_directionality()
# test.create_directionality_dfs()
# test.save_directionality_dfs()
# test.summary_statistics()



import math

import numpy as np
import pandas as pd
from scipy.spatial import ConvexHull
from shapely.affinity import scale
from shapely.geometry import Polygon

from simba.mixins.config_reader import ConfigReader
from simba.mixins.feature_extraction_mixin import FeatureExtractionMixin
from simba.utils.errors import NoROIDataError, ParametersFileError
from simba.utils.printing import stdout_success


class ROISizeStandardizer(ConfigReader, FeatureExtractionMixin):
    """
    Standardize ROI sizes according to a reference video.

    .. note::
       Example: You select a "baseline" video, say that this baseline video has a pixel per millimeter of `10`.
       Say there are a further two videos in the project with ROIs, and these videos has pixels per millimeter of `9` and `11`.
       At runtime, the area of the rectangles, circles and polygons in the two additional videos get their ROI areas increased/decreased
       with 10% while the baseline video ROIs are unchanged.

    :parameter str config_path: path to SimBA project config file in Configparser format.
    :parameter: str join_bouts_within_delta: Name of baseline video without extension, e.g., `Video_1`.

    :example:
    >>> test = ROISizeStandardizer(config_path='/Users/simon/Desktop/envs/troubleshooting/DLC_2_Black_animals/project_folder/project_config.ini', reference_video='Together_1')
    >>> test.run()
    >>> test.save()
    """

    def __init__(self, config_path: str, reference_video: str):
        ConfigReader.__init__(self, config_path=config_path)
        FeatureExtractionMixin.__init__(self)
        self.reference_video_name = reference_video
        self.read_roi_data()
        self.checks()

    def checks(self):
        if self.reference_video_name not in self.video_names_w_rois:
            raise NoROIDataError(
                msg=f"The reference video {self.reference_video_name} does not have any defined ROIs.",
                source=self.__class__.__name__,
            )
        for video_name in self.video_names_w_rois:
            if video_name not in list(self.video_info_df["Video"]):
                raise ParametersFileError(
                    msg=f"Found defined ROIs for video {video_name}, but this video does not have pixels per millimeter defined in the project_folder/logs/video_info.csv file.",
                    source=self.__class__.__name__,
                )

    def compute_correction_factors(self):
        self.correction_factors = {}
        for video_name in self.video_names_w_rois:
            px_per_mm = self.read_video_info(video_name=video_name)[1]
            self.correction_factors[video_name] = round(
                (((px_per_mm / self.reference_px_per_mm) - 1) * 100), 4
            )

    def find_scale_factor(self, video_name: str, correction_factor: float):
        if self.correction_factors[video_name] < 0:
            return math.sqrt(1 - abs(correction_factor) / 100)
        else:
            return math.sqrt(1 + correction_factor / 100)

    def run(self):
        _, self.reference_px_per_mm, _ = self.read_video_info(
            video_name=self.reference_video_name
        )
        self.compute_correction_factors()
        self.video_names_to_change = [
            x for x in self.video_names_w_rois if x != self.reference_video_name
        ]

        updated_rectangles, updated_polygons, updated_circles = (
            pd.DataFrame(columns=self.rectangles_df.columns),
            pd.DataFrame(columns=self.polygon_df.columns),
            pd.DataFrame(columns=self.circles_df.columns),
        )
        for video_name in self.video_names_to_change:
            rectangles = self.rectangles_df[self.rectangles_df["Video"] == video_name]
            circles = self.circles_df[self.circles_df["Video"] == video_name]
            polygons = self.polygon_df[self.polygon_df["Video"] == video_name]

            # UPDATE RECTANGLES
            for idx, rectangle in rectangles.iterrows():
                if self.correction_factors[video_name] != 0:
                    vertices = np.array(
                        [
                            rectangle["Tags"]["Top left tag"],
                            rectangle["Tags"]["Top right tag"],
                            rectangle["Tags"]["Bottom left tag"],
                            rectangle["Tags"]["Bottom right tag"],
                        ]
                    )
                    scale_factor = self.find_scale_factor(
                        video_name=video_name,
                        correction_factor=self.correction_factors[video_name],
                    )
                    new_vertices = np.unique(
                        np.array(
                            scale(
                                Polygon(vertices),
                                xfact=scale_factor,
                                yfact=scale_factor,
                            ).exterior.coords
                        ).astype(np.int16),
                        axis=0,
                    )

                    top_left = np.min(new_vertices, axis=0)
                    bottom_right = np.max(new_vertices, axis=0)

                    rectangle["topLeftX"], rectangle["topLeftY"] = (
                        top_left[0],
                        top_left[1],
                    )
                    rectangle["Bottom_right_X"], rectangle["Bottom_right_Y"] = (
                        bottom_right[0],
                        bottom_right[1],
                    )
                    rectangle["width"], rectangle["height"] = int(
                        rectangle["Bottom_right_X"] - rectangle["topLeftX"]
                    ), int(rectangle["Bottom_right_Y"] - rectangle["topLeftY"])
                    (
                        rectangle["Tags"]["Top left tag"],
                        rectangle["Tags"]["Top right tag"],
                    ) = (int(rectangle["topLeftX"]), int(rectangle["topLeftY"])), (
                        int(rectangle["Bottom_right_X"]),
                        int(rectangle["topLeftY"]),
                    )
                    (
                        rectangle["Tags"]["Bottom left tag"],
                        rectangle["Tags"]["Bottom right tag"],
                    ) = (rectangle["topLeftX"], rectangle["Bottom_right_Y"]), (
                        rectangle["Bottom_right_X"],
                        rectangle["Bottom_right_Y"],
                    )
                    rectangle["Tags"]["Left tag"] = (
                        int(rectangle["topLeftX"]),
                        int(rectangle["topLeftY"] + rectangle["height"] / 2),
                    )
                    rectangle["Tags"]["Right tag"] = (
                        int(rectangle["topLeftX"] + rectangle["width"]),
                        int(rectangle["topLeftY"] + rectangle["height"] / 2),
                    )
                    rectangle["Tags"]["Top tag"] = (
                        int(rectangle["topLeftX"] + rectangle["width"] / 2),
                        int(rectangle["topLeftY"]),
                    )
                    rectangle["Tags"]["Bottom tag"] = (
                        int(rectangle["topLeftX"] + rectangle["width"] / 2),
                        int(rectangle["topLeftY"] + rectangle["height"]),
                    )
                    updated_rectangles.loc[idx] = rectangle

            # UPDATE POLYGONS
            for idx, polygon in polygons.iterrows():
                if self.correction_factors[video_name] != 0:
                    scale_factor = self.find_scale_factor(
                        video_name=video_name,
                        correction_factor=self.correction_factors[video_name],
                    )
                    new_vertices = np.array(
                        scale(
                            Polygon(polygon["vertices"]),
                            xfact=scale_factor,
                            yfact=scale_factor,
                        ).exterior.coords
                    )
                    new_vertices = ConvexHull(new_vertices).points.astype(np.int32)
                    polygon["vertices"] = new_vertices
                    for cnt, i in enumerate(polygon["vertices"]):
                        polygon["Tags"][f"Tag_{cnt}"] = i
                    updated_polygons.loc[idx] = polygon

            # UPDATE CIRCLES
            for idx, circle in circles.iterrows():
                if self.correction_factors[video_name] != 0:
                    circle["radius"] = int(
                        circle["radius"] * (self.correction_factors[video_name] / 100)
                    )
                    circle["Tags"]["Border tag"] = (
                        int(circle["centerX"] - circle["radius"]),
                        circle["centerY"],
                    )
                    updated_circles.loc[idx] = circle

            self.rectangles_df.update(updated_rectangles)
            self.polygon_df.update(updated_polygons)
            self.circles_df.update(updated_circles)

    def save(self):
        store = pd.HDFStore(self.roi_coordinates_path, mode="w")
        store["rectangles"] = self.rectangles_df
        store["circleDf"] = self.circles_df
        store["polygons"] = self.polygon_df
        store.close()
        stdout_success(
            msg=f"ROI size definitions standardized according to pixels per millimeter in video {self.reference_video_name}",
            source=self.__class__.__name__,
        )


# test = ROISizeStandardizer(config_path='/Users/simon/Desktop/envs/troubleshooting/DLC_2_Black_animals/project_folder/project_config.ini', reference_video='Together_1')
# test.run()
# test.save()


#
# test = ROISizeStandardizer(config_path='/Users/simon/Desktop/envs/troubleshooting/two_black_animals_14bp/project_folder/project_config.ini', reference_video='Together_1')
# test.run()
# test.save()



__author__ = "Simon Nilsson"

import itertools
import os

import cv2
import numpy as np

from simba.mixins.config_reader import ConfigReader
from simba.mixins.plotting_mixin import PlottingMixin
from simba.roi_tools.ROI_analyzer import ROIAnalyzer
from simba.utils.data import create_color_palettes
from simba.utils.enums import Formats, Paths
from simba.utils.errors import DuplicationError, NoFilesFoundError
from simba.utils.printing import SimbaTimer, stdout_success
from simba.utils.read_write import get_fn_ext, get_video_meta_data
from simba.utils.warnings import DuplicateNamesWarning


class ROIPlot(ConfigReader, PlottingMixin):
    """
    Visualize the ROI data (number of entries/exits, time-spent in ROIs etc).

    .. note::
       `ROI tutorials <https://github.com/sgoldenlab/simba/blob/master/docs/ROI_tutorial_new.md>`__.

       Use :meth:`simba.plotting.ROI_plotter_mp.ROIPlotMultiprocess` for improved run-time.



    :param str config_path: Path to SimBA project config file in Configparser format
    :param str video_path: Name of video to create ROI visualizations for
    :param dict style_attr: User-defined visualization settings.

    Examples
    ----------
    >>> settings = {'Show_body_part': True, 'Show_animal_name': True}
    >>> roi_visualizer = ROIPlot(ini_path=r'MyProjectConfig', video_path="MyVideo.mp4", settings=settings)
    >>> roi_visualizer.insert_data()
    >>> roi_visualizer.run()
    """

    def __init__(self, ini_path: str, video_path: str, style_attr: dict):
        ConfigReader.__init__(self, config_path=ini_path)
        PlottingMixin.__init__(self)
        self.roi_analyzer = ROIAnalyzer(
            ini_path=ini_path, data_path="outlier_corrected_movement_location"
        )
        self.video_dir_path = os.path.join(self.roi_analyzer.project_path, "videos")
        self.video_path = os.path.join(self.video_dir_path, video_path)
        _, self.video_name, _ = get_fn_ext(video_path)
        self.roi_analyzer.files_found = [
            os.path.join(
                self.roi_analyzer.input_folder,
                self.video_name + "." + self.roi_analyzer.file_type,
            )
        ]
        if not os.path.isfile(self.roi_analyzer.files_found[0]):
            raise NoFilesFoundError(
                msg="SIMBA ERROR: Could not find the file at path {}. Please make sure you have corrected body-part outliers or indicated that you want to skip outlier correction".format(
                    self.roi_analyzer.files_found[0]
                )
            )
        self.roi_analyzer.run()
        self.roi_entries_df = self.roi_analyzer.detailed_df
        self.data_df, self.style_attr = self.roi_analyzer.data_df, style_attr
        self.video_shapes = list(
            itertools.chain(
                self.roi_analyzer.video_recs["Name"].unique(),
                self.roi_analyzer.video_circs["Name"].unique(),
                self.roi_analyzer.video_polys["Name"].unique(),
            )
        )
        if len(list(set(self.video_shapes))) != len(self.video_shapes):
            raise DuplicationError(
                msg="Some SIMBA ROI shapes have identical names. Please use unique names to visualize ROI data."
            )
        self.roi_analyzer.video_recs = self.add_missing_ROI_cols(
            self.roi_analyzer.video_recs
        )
        self.roi_analyzer.video_circs = self.add_missing_ROI_cols(
            self.roi_analyzer.video_circs
        )
        self.roi_analyzer.video_polys = self.add_missing_ROI_cols(
            self.roi_analyzer.video_polys
        )

        self.shape_dicts = {}
        for df in [
            self.roi_analyzer.video_recs,
            self.roi_analyzer.video_circs,
            self.roi_analyzer.video_polys,
        ]:
            if not df["Name"].is_unique:
                df = df.drop_duplicates(subset=["Name"], keep="first")
                DuplicateNamesWarning(
                    msg='Some of your ROIs with the same shape has the same names. E.g., you have two rectangles named "My rectangle". SimBA prefers ROI shapes with unique names. SimBA will keep one of the unique shape names and drop the rest.'
                )
            d = df.set_index("Name").to_dict(orient="index")
            self.shape_dicts = {**self.shape_dicts, **d}

        self.shape_columns = []
        for animal in self.roi_analyzer.multi_animal_id_list:
            for shape_name in self.video_shapes:
                self.data_df[animal + "_" + shape_name] = 0
                self.shape_columns.append(animal + "_" + shape_name)
        self.bp_dict = self.roi_analyzer.bp_dict
        self.output_folder = os.path.join(
            self.roi_analyzer.project_path, Paths.ROI_ANALYSIS.value
        )
        if not os.path.exists(self.output_folder):
            os.makedirs(self.output_folder)
        self.timer = SimbaTimer()
        self.timer.start_timer()

    def insert_data(self):
        """
        Method to concat ROI and pose-estimation data.

        Returns
        -------
        Attribute: pd.Dataframe
            data_df

        """
        self.roi_entries_dict = self.roi_entries_df[
            ["ANIMAL", "SHAPE", "ENTRY FRAMES", "EXIT FRAMES"]
        ].to_dict(orient="records")
        for entry_dict in self.roi_entries_dict:
            entry, exit = int(entry_dict["ENTRY FRAMES"]), int(
                entry_dict["EXIT FRAMES"]
            )
            entry_dict["frame_range"] = list(range(entry, exit + 1))
            col_name = entry_dict["ANIMAL"] + "_" + entry_dict["SHAPE"]
            self.data_df[col_name][
                self.data_df.index.isin(entry_dict["frame_range"])
            ] = 1

    def __calc_text_locs(self):
        add_spacer = 2
        self.loc_dict = {}
        for animal_cnt, animal_name in enumerate(
            self.roi_analyzer.multi_animal_id_list
        ):
            self.loc_dict[animal_name] = {}
            for shape in self.video_shapes:
                self.loc_dict[animal_name][shape] = {}
                self.loc_dict[animal_name][shape]["timer_text"] = "{} {} {}".format(
                    shape, animal_name, "timer:"
                )
                self.loc_dict[animal_name][shape]["entries_text"] = "{} {} {}".format(
                    shape, animal_name, "entries:"
                )
                self.loc_dict[animal_name][shape]["timer_text_loc"] = (
                    (self.video_meta_data["width"] + 5),
                    (
                        self.video_meta_data["height"]
                        - (self.video_meta_data["height"] + 10)
                        + self.spacing_scaler * add_spacer
                    ),
                )
                self.loc_dict[animal_name][shape]["timer_data_loc"] = (
                    int(self.border_img_w - (self.border_img_w / 8)),
                    (
                        self.video_meta_data["height"]
                        - (self.video_meta_data["height"] + 10)
                        + self.spacing_scaler * add_spacer
                    ),
                )
                add_spacer += 1
                self.loc_dict[animal_name][shape]["entries_text_loc"] = (
                    (self.video_meta_data["width"] + 5),
                    (
                        self.video_meta_data["height"]
                        - (self.video_meta_data["height"] + 10)
                        + self.spacing_scaler * add_spacer
                    ),
                )
                self.loc_dict[animal_name][shape]["entries_data_loc"] = (
                    int(self.border_img_w - (self.border_img_w / 8)),
                    (
                        self.video_meta_data["height"]
                        - (self.video_meta_data["height"] + 10)
                        + self.spacing_scaler * add_spacer
                    ),
                )
                add_spacer += 1

    def __insert_texts(self, shape_df):
        for animal_name in self.roi_analyzer.multi_animal_id_list:
            for _, shape in shape_df.iterrows():
                shape_name, shape_color = shape["Name"], shape["Color BGR"]
                cv2.putText(
                    self.border_img,
                    self.loc_dict[animal_name][shape_name]["timer_text"],
                    self.loc_dict[animal_name][shape_name]["timer_text_loc"],
                    self.font,
                    self.font_size,
                    shape_color,
                    1,
                )
                cv2.putText(
                    self.border_img,
                    self.loc_dict[animal_name][shape_name]["entries_text"],
                    self.loc_dict[animal_name][shape_name]["entries_text_loc"],
                    self.font,
                    self.font_size,
                    shape_color,
                    1,
                )

    def __create_counters(self):
        self.cnt_dict = {}
        for animal_cnt, animal_name in enumerate(
            self.roi_analyzer.multi_animal_id_list
        ):
            self.cnt_dict[animal_name] = {}
            for shape in self.video_shapes:
                self.cnt_dict[animal_name][shape] = {}
                self.cnt_dict[animal_name][shape]["timer"] = 0
                self.cnt_dict[animal_name][shape]["entries"] = 0
                self.cnt_dict[animal_name][shape]["entry_status"] = False

    def __update_video_meta_data(self):
        new_cap = cv2.VideoCapture(self.video_path)
        new_cap.set(1, 1)
        _, img = self.cap.read()
        bordered_img = cv2.copyMakeBorder(
            img,
            0,
            0,
            0,
            int(self.video_meta_data["width"]),
            borderType=cv2.BORDER_CONSTANT,
            value=[0, 0, 0],
        )
        self.border_img_h, self.border_img_w = (
            bordered_img.shape[0],
            bordered_img.shape[1],
        )
        new_cap.release()

    def visualize_ROI_data(self):
        """
        Method to visualize ROI data. Results are stored in the `project_folder/frames/output/ROI_analysis`
        directroy of the SimBA project.

        Returns
        -------
        None
        """

        save_path = os.path.join(self.output_folder, self.video_name + ".avi")
        self.cap = cv2.VideoCapture(self.video_path)
        self.video_meta_data = get_video_meta_data(self.video_path)
        video_settings, pix_per_mm, fps = self.read_video_info(
            video_name=self.video_name
        )
        self.space_scale, radius_scale, res_scale, font_scale = 25, 10, 1500, 0.8
        max_dim = max(self.video_meta_data["width"], self.video_meta_data["height"])
        self.font = cv2.FONT_HERSHEY_TRIPLEX
        draw_scale, self.font_size = int(radius_scale / (res_scale / max_dim)), float(
            font_scale / (res_scale / max_dim)
        )
        fourcc = cv2.VideoWriter_fourcc(*Formats.AVI_CODEC.value)
        self.spacing_scaler = int(self.space_scale / (res_scale / max_dim))
        writer = cv2.VideoWriter(
            save_path,
            fourcc,
            fps,
            (self.video_meta_data["width"] * 2, self.video_meta_data["height"]),
        )
        color_lst = create_color_palettes(
            self.roi_analyzer.animal_cnt, int((len(self.roi_analyzer.bp_names) / 3))
        )[0]
        self.__update_video_meta_data()
        self.__calc_text_locs()
        self.__create_counters()

        frame_cnt = 0
        while self.cap.isOpened():
            ret, img = self.cap.read()
            try:
                if ret:
                    self.border_img = cv2.copyMakeBorder(
                        img,
                        0,
                        0,
                        0,
                        int(self.video_meta_data["width"]),
                        borderType=cv2.BORDER_CONSTANT,
                        value=[0, 0, 0],
                    )
                    self.border_img_h, self.border_img_w = (
                        self.border_img.shape[0],
                        self.border_img.shape[1],
                    )
                    self.__insert_texts(self.roi_analyzer.video_recs)
                    self.__insert_texts(self.roi_analyzer.video_circs)
                    self.__insert_texts(self.roi_analyzer.video_polys)

                    for _, row in self.roi_analyzer.video_recs.iterrows():
                        top_left_x, top_left_y, shape_name = (
                            row["topLeftX"],
                            row["topLeftY"],
                            row["Name"],
                        )
                        bottom_right_x, bottom_right_y = (
                            row["Bottom_right_X"],
                            row["Bottom_right_Y"],
                        )
                        thickness, color = row["Thickness"], row["Color BGR"]
                        cv2.rectangle(
                            self.border_img,
                            (top_left_x, top_left_y),
                            (bottom_right_x, bottom_right_y),
                            color,
                            thickness,
                        )

                    for _, row in self.roi_analyzer.video_circs.iterrows():
                        center_x, center_y, radius, shape_name = (
                            row["centerX"],
                            row["centerY"],
                            row["radius"],
                            row["Name"],
                        )
                        thickness, color = row["Thickness"], row["Color BGR"]
                        cv2.circle(
                            self.border_img,
                            (center_x, center_y),
                            radius,
                            color,
                            thickness,
                        )

                    for _, row in self.roi_analyzer.video_polys.iterrows():
                        vertices, shape_name = row["vertices"], row["Name"]
                        thickness, color = row["Thickness"], row["Color BGR"]
                        cv2.polylines(
                            self.border_img,
                            [vertices],
                            True,
                            color,
                            thickness=thickness,
                        )

                    for animal_cnt, animal_name in enumerate(
                        self.roi_analyzer.multi_animal_id_list
                    ):
                        bp_data = self.data_df.loc[
                            frame_cnt, self.bp_dict[animal_name]
                        ].values
                        if self.roi_analyzer.settings["threshold"] < bp_data[2]:
                            if self.style_attr["Show_body_part"]:
                                cv2.circle(
                                    self.border_img,
                                    (int(bp_data[0]), int(bp_data[1])),
                                    draw_scale,
                                    color_lst[animal_cnt],
                                    -1,
                                )
                            if self.style_attr["Show_animal_name"]:
                                cv2.putText(
                                    self.border_img,
                                    animal_name,
                                    (int(bp_data[0]), int(bp_data[1])),
                                    self.font,
                                    self.font_size,
                                    color_lst[animal_cnt],
                                    1,
                                )
                            for shape_name in self.video_shapes:
                                if (
                                    self.data_df.loc[
                                        frame_cnt, animal_name + "_" + shape_name
                                    ]
                                    == 1
                                ):
                                    self.cnt_dict[animal_name][shape_name]["timer"] += (
                                        1 / fps
                                    )
                                    if not self.cnt_dict[animal_name][shape_name][
                                        "entry_status"
                                    ]:
                                        self.cnt_dict[animal_name][shape_name][
                                            "entry_status"
                                        ] = True
                                        self.cnt_dict[animal_name][shape_name][
                                            "entries"
                                        ] += 1
                                else:
                                    self.cnt_dict[animal_name][shape_name][
                                        "entry_status"
                                    ] = False

                                cv2.putText(
                                    self.border_img,
                                    str(
                                        round(
                                            self.cnt_dict[animal_name][shape_name][
                                                "timer"
                                            ],
                                            2,
                                        )
                                    ),
                                    self.loc_dict[animal_name][shape_name][
                                        "timer_data_loc"
                                    ],
                                    self.font,
                                    self.font_size,
                                    self.shape_dicts[shape_name]["Color BGR"],
                                    1,
                                )
                                cv2.putText(
                                    self.border_img,
                                    str(
                                        self.cnt_dict[animal_name][shape_name][
                                            "entries"
                                        ]
                                    ),
                                    self.loc_dict[animal_name][shape_name][
                                        "entries_data_loc"
                                    ],
                                    self.font,
                                    self.font_size,
                                    self.shape_dicts[shape_name]["Color BGR"],
                                    1,
                                )

                    writer.write(np.uint8(self.border_img))
                    # cv2.imshow('Window', self.border_img)
                    # key = cv2.waitKey(3000)
                    # if key == 27:
                    #     cv2.destroyAllWindows()
                    print(
                        "Frame: {} / {}, Video: {}.".format(
                            str(frame_cnt),
                            str(self.video_meta_data["frame_count"]),
                            self.video_name,
                        )
                    )
                    frame_cnt += 1

                if img is None:
                    self.timer.stop_timer()
                    print(
                        f"SIMBA COMPLETE: Video {self.video_name} saved at {save_path} (elapsed time: {self.timer.elapsed_time_str}s)."
                    )
                    self.cap.release()
                    break

            except Exception as e:
                writer.release()
                print(e.args)
                print(
                    "NOTE: index error / keyerror. Some frames of the video may be missing. Make sure you are running latest version of SimBA with pip install simba-uw-tf-dev"
                )
                break

        writer.release()
        stdout_success(
            msg=f"Video {self.video_name} created. Video saved in project_folder/frames/output/ROI_analysis"
        )


# test = ROIPlot(ini_path=r'/Users/simon/Desktop/envs/troubleshooting/Termites_5/project_folder/project_config.ini',
#                video_path="termite_test.mp4",
#                style_attr={'Show_body_part': True, 'Show_animal_name': True})
# test.insert_data()
# test.visualize_ROI_data()


# test = ROIPlot(ini_path=r'/Users/simon/Desktop/envs/troubleshooting/Termites_5/project_folder/project_config.ini', video_path="termite_test.mp4")
# test.insert_data()
# test.visualize_ROI_data()

# test = ROIPlot(ini_path=r'/Users/simon/Desktop/troubleshooting/train_model_project/project_folder/project_config.ini', video_path=r"Together_1.avi")
# test.insert_data()
# test.visualize_ROI_data()


# test = ROIPlot(ini_path=r'/Users/simon/Desktop/envs/troubleshooting/two_black_animals_14bp/project_folder/project_config.ini',
#                video_path="Together_2.avi",
#                style_attr={'Show_body_part': True, 'Show_animal_name': False})
# test.insert_data()
# test.visualize_ROI_data()



__author__ = "Simon Nilsson"

import glob
import json
import os
from copy import deepcopy
from typing import Union

import numpy as np
import pandas as pd
import pyarrow as pa
import pyarrow.parquet as pq

from simba.data_processors.interpolate_pose import Interpolate
from simba.utils.data import smooth_data_gaussian, smooth_data_savitzky_golay
from simba.utils.enums import Dtypes, Methods, Paths
from simba.utils.errors import NoFilesFoundError
from simba.utils.printing import stdout_success
from simba.utils.read_write import (get_fn_ext, read_config_file,
                                    read_project_path_and_file_type, write_df)


class MarsImporter(object):
    """
    Import two animal MARS pose-estimation data (in JSON format) into a SimBA project in
    parquet or CSV format.

    :param str config_path: path to SimBA project config file in Configparser format
    :param str data_folder: Path to file (or directory) with data in `.json` format.
    :param str interpolation_settings: String defining the pose-estimation interpolation method. OPTIONS: 'None', 'Animal(s): Nearest',
        'Animal(s): Linear', 'Animal(s): Quadratic','Body-parts: Nearest', 'Body-parts: Linear',
        'Body-parts: Quadratic'. See `this image <https://gist.github.com/gyassine/b47b90e8de935cc06ef856401f7582b0>`>_ for comparison of interpolation methods.
    :param dict smoothing_method: Dictionary defining the pose estimation smoothing method. EXAMPLE: {'Method': 'Savitzky Golay',
        'Parameters': {'Time_window': '200'}})

    .. note::
       `Multi-animal import tutorial <https://github.com/sgoldenlab/simba/blob/master/docs/Multi_animal_pose.md>`__.

    Examples
    -----
    >>> mars_importer = MarsImporter(config_path=r'MyConfigPath', data_folder=r'MyMarsDataFolder', interpolation_settings='None', smoothing_settings={'Method': 'None', 'Parameters': {'Time_window': '200'}})
    >>> mars_importer.import_data()

    References
    ----------
    .. [1] Segalin et al., The Mouse Action Recognition System (MARS) software pipeline for automated analysis of social behaviors in mice, `eLife`, 2021.

    """

    def __init__(
        self,
        config_path: Union[str, os.PathLike],
        data_path: Union[str, os.PathLike],
        interpolation_method: str,
        smoothing_method: dict,
    ):
        self.config, self._config_path = (
            read_config_file(config_path=config_path),
            config_path,
        )
        self.data_path = data_path
        self.interpolation_method, self.smoothing_method = (
            interpolation_method,
            smoothing_method,
        )
        self.project_path, self.file_type = read_project_path_and_file_type(
            config=self.config
        )
        self.save_dir = os.path.join(self.project_path, Paths.INPUT_CSV.value)
        if os.path.isdir(data_path):
            self.files_found = glob.glob(data_path + "/*.json")
        else:
            self.files_found = [data_path]
        if len(self.files_found) == 0:
            raise NoFilesFoundError(
                msg=f"Zero .json files found in {data_path} directory"
            )
        body_part_names = [
            "Nose",
            "Ear_left",
            "Ear_right",
            "Neck",
            "Hip_left",
            "Hip_right",
            "Tail",
        ]
        self.keypoint_headers, self.scores_headers = [], []
        for animal in ["1", "2"]:
            for body_part in body_part_names:
                for coordinate in ["x", "y"]:
                    self.keypoint_headers.append(
                        body_part + "_" + animal + "_" + coordinate
                    )
        for animal in ["1", "2"]:
            for body_part in body_part_names:
                self.scores_headers.append(body_part + "_" + animal + "_p")
        self.headers = deepcopy(self.keypoint_headers)
        index = 3 - 1
        for elem in self.scores_headers:
            self.headers.insert(index, elem)
            index += 3

    def __merge_dfs(self, df_1: pd.DataFrame, df_2: pd.DataFrame):
        df = pd.DataFrame()
        for cnt, c in enumerate(df_1.columns):
            df[len(df.columns)] = df_1[c]
            df[len(df.columns)] = df_2[c]
        return df

    def __create_multi_index_headers(self, df: pd.DataFrame):
        multi_index_tuples = []
        for column in range(len(df.columns)):
            multi_index_tuples.append(tuple(("MARS", "MARS", df.columns[column])))
        df.columns = pd.MultiIndex.from_tuples(
            multi_index_tuples, names=["scorer", "bodypart", "coords"]
        )
        return df

    def __perform_interpolation(
        self,
        file_path: str,
        workflow_file_type: str,
        config_path: str,
        interpolation_method: str,
    ):
        if workflow_file_type == "parquet":
            df = pd.read_parquet(file_path)
        else:
            df = pd.read_csv(file_path, index_col=0)
        interpolate_body_parts = Interpolate(config_path, df)
        interpolate_body_parts.detect_headers()
        interpolate_body_parts.fix_missing_values(interpolation_method)
        interpolate_body_parts.reorganize_headers()
        if workflow_file_type == "parquet":
            table = pa.Table.from_pandas(interpolate_body_parts.new_df)
            pq.write_table(table, file_path)
        if workflow_file_type == "csv":
            interpolate_body_parts.new_df.to_csv(file_path)

    def __run_smoothing(self):
        if self.smoothing_method["Method"] == Methods.GAUSSIAN.value:
            print("Performing Gaussian smoothing on video {}...".format(self.file_name))
            time_window = self.smoothing_method["Parameters"]["Time_window"]
            smooth_data_gaussian(
                config=self.config,
                file_path=self.save_path,
                time_window_parameter=time_window,
            )

        if self.smoothing_method["Method"] == Methods.SAVITZKY_GOLAY.value:
            print(
                "Performing Savitzky Golay smoothing on video {}...".format(
                    self.file_name
                )
            )
            time_window = self.smoothing_method["Parameters"]["Time_window"]
            smooth_data_savitzky_golay(
                config=self.config,
                file_path=self.save_path,
                time_window_parameter=time_window,
            )

    def import_data(self):
        for file_cnt, file_path in enumerate(self.files_found):
            _, self.file_name, _ = get_fn_ext(file_path)
            print("Importing data for video {}...".format(self.file_name))
            self.save_path = os.path.join(
                self.save_dir, self.file_name + "." + self.file_type
            )
            with open(file_path, "r") as j:
                data = json.loads(j.read())
            key_points, scores = np.array(data["keypoints"]).astype(int), np.array(
                data["scores"]
            )
            animal_1_scores, animal_2_scores = pd.DataFrame(scores[:, 0]), pd.DataFrame(
                scores[:, 1]
            )
            data_df = []
            for a in [key_points[:, 0], key_points[:, 1]]:
                m, n, r = a.shape
                arr = np.column_stack(
                    (np.repeat(np.arange(m), n), a.reshape(m * n, -1))
                )
                df = pd.DataFrame(arr)
                df_x, df_y = df[df.index % 2 != 0].set_index(0), df[
                    df.index % 2 != 1
                ].set_index(0)
                data_df.append(self.__merge_dfs(df_x, df_y))
            data_df = pd.concat(data_df, axis=1)
            data_df.columns = self.keypoint_headers
            scores_df = pd.concat([animal_1_scores, animal_2_scores], axis=1)
            scores_df.columns = self.scores_headers
            data_df = pd.concat([data_df, scores_df], axis=1)[self.headers]
            data_df = self.__create_multi_index_headers(df=data_df)
            write_df(data_df, self.file_type, self.save_path)

            if self.interpolation_method != Dtypes.NONE.value:
                print("Performing interpolation...")
                self.__perform_interpolation(
                    self.save_path,
                    self.file_type,
                    self._config_path,
                    self.interpolation_method,
                )
            if self.smoothing_method["Method"] != Dtypes.NONE.value:
                self.__run_smoothing()
            print("Video imported {}.".format(self.file_name))
        stdout_success(
            msg=f"{str(len(self.files_found))} data files imported to SimBA project"
        )



__author__ = "Simon Nilsson"

import glob
import os
from typing import List, Union

import cv2
import imutils
import numpy as np

import simba
from simba.mixins.plotting_mixin import PlottingMixin
from simba.utils.checks import check_file_exist_and_readable
from simba.utils.data import create_color_palettes
from simba.utils.enums import Paths
from simba.utils.errors import InvalidFileTypeError


class PoseConfigCreator(PlottingMixin):
    """

    Class for creating user-defined pose-estimation pipeline in SimBA through a GUI interface.

    ..seealso::
       `GitHub tutorial/documentation <https://github.com/sgoldenlab/simba/blob/master/docs/Pose_config.md>`__.

    :param str pose_name: Name of the user-defined pose-estimation setting.
    :param str no_animals: Number of animals in the user-defined pose-estimation setting.
    :param str img_path: Path to image representation of user-defined pose-estimation setting
    :param List[str] bp_list: Body-parts in the user-defined pose-estimation setting.
    :param List[int] animal_id_int_list: Integers representing the animal ID which the body-parts belong to.

    :examples:
    >>> pose_config_creator = PoseConfigCreator(pose_name="My_test_config", no_animals=2, img_path='simba/splash_050122.png', bp_list=['Ear', 'Nose', 'Left_ear', 'Ear', 'Nose', 'Left_ear'], animal_id_int_list= [1, 1, 1, 2, 2, 2])
    >>> pose_config_creator.launch()
    """

    def __init__(
        self,
        pose_name: str,
        no_animals: int,
        img_path: Union[str, os.PathLike],
        bp_list: List[str],
        animal_id_int_list: List[str],
    ):
        PlottingMixin.__init__(self)
        self.pose_name, self.img_path = pose_name, img_path
        self.bp_list, self.animal_id_int_list = bp_list, animal_id_int_list
        check_file_exist_and_readable(img_path)
        self.no_animals, self.font = no_animals, cv2.FONT_HERSHEY_SIMPLEX
        self.img = cv2.imread(img_path)
        if not isinstance(self.img, (list, tuple, np.ndarray)):
            raise InvalidFileTypeError(
                msg=f"The chosen image file could not be read as an image ({img_path})"
            )
        self.img_height, self.img_width = int(self.img.shape[0]), int(self.img.shape[1])
        if self.img_width < 300:
            self.img = imutils.resize(self.img, width=800)
            self.img_height, self.img_width = int(self.img.shape[0]), int(
                self.img.shape[1]
            )
            self.img = np.uint8(self.img)
        self.space_scale, self.radius_scale, self.res_scale, self.font_scale = (
            60,
            12,
            1500,
            1.5,
        )
        self.max_dim = max(self.img_width, self.img_height)
        self.circle_scale = int(self.radius_scale / (self.res_scale / self.max_dim))
        self.font_size = float(self.font_scale / (self.res_scale / self.max_dim))
        self.spacing_scale = int(self.space_scale / (self.res_scale / self.max_dim))
        cv2.namedWindow("Define pose", cv2.WINDOW_NORMAL)
        self.overlay = self.img.copy()

        if self.no_animals > 1:
            for cnt, (bp_name, animal_number_id) in enumerate(
                zip(self.bp_list, self.animal_id_int_list)
            ):
                self.bp_list[cnt] = "{}_{}".format(bp_name, animal_number_id)
        self.color_lst = create_color_palettes(1, len(self.bp_list))[0]

    def launch(self):
        def draw_circle(event, x, y, flags, param):
            if event == cv2.EVENT_LBUTTONDBLCLK:
                cv2.circle(
                    self.overlay,
                    (x, int(y - self.side_img.shape[0])),
                    self.circle_scale,
                    self.color_lst[self.bp_cnt],
                    -1,
                )
                cv2.putText(
                    self.overlay,
                    str(self.bp_cnt + 1),
                    (x + 4, int(y - self.side_img.shape[0])),
                    cv2.FONT_HERSHEY_SIMPLEX,
                    self.font_size,
                    self.color_lst[self.bp_cnt],
                    2,
                )
                self.cord_written = True

        for bp_cnt, bp_name in enumerate(self.bp_list):
            self.cord_written = False
            self.bp_cnt = bp_cnt
            self.side_img = np.zeros(
                (int(self.img_height / 4), self.img_width, 3), np.uint8
            )
            cv2.putText(
                self.side_img,
                "Double left click on body part {}.".format(bp_name),
                (10, 50),
                self.font,
                self.font_size,
                self.color_lst[bp_cnt],
                2,
            )
            img_concat = cv2.vconcat([self.side_img, self.overlay])
            cv2.namedWindow("Define pose", cv2.WINDOW_NORMAL)
            cv2.imshow("Define pose", img_concat)
            while not self.cord_written:
                cv2.setMouseCallback("Define pose", draw_circle)
                img_concat = cv2.vconcat([self.side_img, self.overlay])
                cv2.namedWindow("Define pose", cv2.WINDOW_NORMAL)
                cv2.imshow("Define pose", img_concat)
                cv2.waitKey(1)
            cv2.destroyWindow("Define pose")
        self.save()

    def save(self):
        overlay = cv2.resize(self.overlay, (250, 300))
        simba_cw = os.path.dirname(simba.__file__)
        img_dir = os.path.join(simba_cw, Paths.SCHEMATICS.value)
        pose_name_path = os.path.join(simba_cw, Paths.PROJECT_POSE_CONFIG_NAMES.value)
        bp_path = os.path.join(simba_cw, Paths.SIMBA_BP_CONFIG_PATH.value)
        no_animals_path = os.path.join(simba_cw, Paths.SIMBA_NO_ANIMALS_PATH.value)
        for path in [pose_name_path, bp_path, no_animals_path]:
            check_file_exist_and_readable(file_path=path)
        prior_imgs = len(glob.glob(img_dir + "/*.png"))
        new_img_name = f"{str(prior_imgs+1)}.png"
        new_img_path = os.path.join(img_dir, new_img_name)
        self.bp_list = ",".join(self.bp_list)
        with open(pose_name_path, "a") as f:
            f.write(self.pose_name + "\n")
        with open(bp_path, "a") as fd:
            fd.write(self.bp_list + "\n")
        with open(no_animals_path, "a") as fd:
            fd.write(str(self.no_animals) + "\n")
        cv2.imwrite(new_img_path, overlay)


# pose_config_creator = PoseConfigCreator(pose_name="My_test_config", no_animals=1, img_path='/Users/simon/Desktop/rotate.png', bp_list=['Ear', 'Nose', 'Left_ear'], animal_id_int_list= [1, 1, 1, 2, 2, 2])
# pose_config_creator.launch()
